﻿// Shawn Poulson, http://explodingcoder.com
// 2010.02.04
//
// Released under Creative Commons Attribution-Share Alike license
// http://creativecommons.org/licenses/by-sa/2.5/

using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Threading;
using Expl.Auxiliary;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Memoize_Test {
   /// <summary>
   /// Unit tests against Memoize with ValidationDictionary.
   /// </summary>
   [TestClass]
   public class MemoizeValidationTest {
      /// <summary>
      /// Memoize a function and apply a predicate that always passes.
      /// </summary>
      [TestMethod]
      public void AlwaysValidTest() {
         var dict = new ValidationDictionary<string, string>(x => true);
         var counter = 0;
         var func = new Func<string, string>(delegate(string x) {
            counter++;
            return x.GetHashCode().ToString();
         });
         var memoizedFunc = Cache.Memoize(func, dict);

         // Memoize keeps everything valid unconditionally
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(5, counter, "Function counter mismatch");
      }

      /// <summary>
      /// Memoize a function and apply a predicate that always fails.
      /// </summary>
      [TestMethod]
      public void AlwaysInvalidTest() {
         var dict = new ValidationDictionary<string, string>(x => false);
         var counter = 0;
         var func = new Func<string, string>(delegate(string x) {
            counter++;
            return x.GetHashCode().ToString();
         });
         var memoizedFunc = Cache.Memoize(func, dict);

         // Everything always invalidated
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(8, counter, "Function counter mismatch");
      }

      /// <summary>
      /// Memoize a function and apply a global timeout that expires in a short amount of time.
      /// </summary>
      [TestMethod]
      public void GlobalTimeoutValidTest() {
         var timeoutDuration = TimeSpan.FromMilliseconds(250);
         var expirationTime = DateTime.Now + timeoutDuration;
         var dict = new ValidationDictionary<string, string>(x => DateTime.Now < expirationTime);
         var counter = 0;
         var func = new Func<string, string>(delegate(string x) {
            counter++;
            return x.GetHashCode().ToString();
         });
         var memoizedFunc = Cache.Memoize(func, dict);

         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter++
         Assert.AreEqual(5, counter, "Function counter mismatch");

         // Sleep until timeout
         Thread.Sleep(timeoutDuration.Add(TimeSpan.FromMilliseconds(250))); // buffer add'l 250ms

         // Timeout expired, so everything stays invalidated
         counter = 0;
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(8, counter, "Function counter mismatch");
      }

      /// <summary>
      /// Memoize a function and apply a global timeout that expires immediately.
      /// </summary>
      [TestMethod]
      public void GlobalTimeoutInvalidTest() {
         var expirationTime = DateTime.Now;
         var dict = new ValidationDictionary<string, string>(x => DateTime.Now < expirationTime);
         var counter = 0;
         var func = new Func<string, string>(delegate(string x) {
            counter++;
            return x.GetHashCode().ToString();
         });
         var memoizedFunc = Cache.Memoize(func, dict);

         // Timeout expires immediately, so everything stays invalidated
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(func("a"), memoizedFunc("a"), "Function value mismatch"); // counter += 2
         Assert.AreEqual(8, counter, "Function counter mismatch");
      }

      /// <summary>
      /// Memoize a function and apply a per-value timeout.
      /// This uses the Timeoutable generic as a value type to assign each
      /// value a timeout and perform validation against it.
      /// </summary>
      [TestMethod]
      public void TimeoutableTest() {
         var timeoutDuration = TimeSpan.FromMilliseconds(250);
         ValidationDictionary<string, Timeoutable<string>> dict = null;
         dict = new ValidationDictionary<string, Timeoutable<string>>(pair => pair.Value.IsValid());
         var counter = 0;
         var func = new Func<string, Timeoutable<string>>(delegate(string x) {
            counter++;
            return new Timeoutable<string>(x.GetHashCode().ToString(), timeoutDuration);
         });
         var memoizedFunc = Cache.Memoize<string, Timeoutable<string>>(func, dict);

         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(5, counter, "Function counter mismatch");

         // Sleep until timeout
         Thread.Sleep(timeoutDuration.Add(TimeSpan.FromMilliseconds(250))); // buffer add'l 250ms

         // Timeout expired, so next call is invalidated, timeout reset
         counter = 0;
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter += 2
         Assert.AreEqual(2, counter, "Function counter mismatch");
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(func("a").Value, memoizedFunc("a").Value, "Function value mismatch"); // counter++
         Assert.AreEqual(5, counter, "Function counter mismatch");
      }

      /// <summary>
      /// Memoize a web request with timeout
      /// </summary>
      [TestMethod]
      public void WebRequestTimeoutTest() {
         var timeoutDuration = TimeSpan.FromMilliseconds(500);
         var dict = new ValidationDictionary<Uri, IValidatable<string>>(pair => pair.Value.IsValid());
         var counter = 0;

         var func = new Func<Uri, IValidatable<string>>(delegate(Uri requestUri) {
            var response = (HttpWebResponse)HttpWebRequest.Create(requestUri).GetResponse();
            var content = new StreamReader(response.GetResponseStream()).ReadToEnd();
            counter++;
            return new Timeoutable<string>(content, timeoutDuration);
         });

         var memoizedFunc = Cache.Memoize(func, dict);

         var urlList = new[] {
            new Uri("http://feeds.digg.com/digg/popular.rss"),
            new Uri("http://stackoverflow.com/feeds"),
            new Uri("http://news.google.com/news?pz=1&cf=all&ned=us&hl=en&topic=h&num=3&output=rss")
         };

         foreach (var url in urlList) {
            string content1 = string.Empty;
            counter = 0;

            // Call web request a few times, should make only one actual web request
            for (int i = 0; i < 3; i++) {
               content1 = memoizedFunc(url).Value;
            }
            Assert.AreEqual(1, counter, "Function counter mismatch");

            // Sleep until timeout
            Thread.Sleep(timeoutDuration + TimeSpan.FromMilliseconds(250)); // buffer add'l 250ms

            // Timeout expired, so next call is invalidated, timeout reset
            counter = 0;
            for (int i = 0; i < 3; i++) {
               var content2 = memoizedFunc(url);
            }
            Assert.AreEqual(1, counter, "Function counter mismatch");

            // Show response content
            Debug.WriteLine(string.Format("Request Uri: {0}", url));
            Debug.WriteLine(string.Format("Content: ({1} bytes){0}{2}...{0}---{0}{0}", Environment.NewLine, content1.Length, content1.Substring(0, Math.Min(content1.Length, 1024))));
         }
      }
   }
}
